##############################################################################
# ti92.tcl
#########################
#
#   Copyright (C) David Kuehling 1999/2000
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

# This file provides commands for link transfer between a computer and a
# TI-92 calculator. It might be possible to make this TI-92+ compatible, 
# however I do not own a TI-92+, so I haven't tried...
#
# The transfer routines currently only support the file format used by TI-
# Graph Link (*.92?). Neither group-files (*.92g) nor *.92@ are supported
# (yet?). If you want to use these routines, you first have to load this Tcl 
# script by the 'source' command, e.g:
#
#     set libdir "/usr/local/lib";   # or any other directory...
#     source "$libdir/ti92.tcl";     # load TI-92 link transfer commands
#
# This will create a namespace 'TI92' which exports the routines 'OpenPort',
# 'ClosePort', 'CheckLink', 'ReceiveFile92', 'ReceiveScreen', 'SendFile92', 
# 'ReceiveBackup', 'SendBackup', 'ReceiveFS' and 'SendKey'. You may then 
# access those commands by using the qualified namespace name, e.g:
#
#     set device "/dev/ttyS2";          # use instead com1...com4 in dos/win95
#     TI92::OpenPort $device;           # open port
#     TI92::CheckLink;                  # check whether calc is available
#     TI92::SendBackup "backup.92b" ""; # send some backup...
#
# Or you may import those commands by the 'namespace import' command, so you 
# can just write 'CheckLink' instead of 'TI92::CheckLink':
#
#     namespace import "TI92::*";    # import link commands
#     set device "/dev/ttyS2";       # use instead com1...com4 in dos/win95
#     OpenPort $device;              # open port
#     CheckLink;                     # check whether calc is available
#     SendBackup "backup.92b" "";    # send some backup...
#
# Note that none of those commands will work with a Tcl version earlier than
# 8.0. I also experienced problems in the Tcl 8.0 linux version, updating to 
# 8.2 fixed the problem. I recommend you use Tcl 8.2.
#
# All the commands of 'ti92.tcl' will generate errors if something went wrong
# (e.g. timeout error, I/O exception, checksum error etc.). So you should use
# 'catch' for error detection.
#
# Command description:
#============================================================================
#
# NAME
#       TI92::OpenPort
# SYNOPSIS
#       TI92::OpenPort device
# DESCRIPTION
#       Open the serial device 'device' for link transfer to and from a 
#       TI-92. This routine sets up some global variables, so it has to be
#       called before using any other link routines. In dos/win95 'device'
#       should be "comX" where X is 1...4. In linux it should refer to any
#       device-file which is attached to a serial port. (e.g. "/dev/ttyS0").
#
#---------------------------------------------------------------------------
#
# NAME 
#       TI92::ClosePort
# SYNOPSIS
#       TI92::ClosePort
# DESCRIPTION
#       Close the device which was previously opened by `TI92::OpenPort'.
#
#---------------------------------------------------------------------------
#
# NAME
#       TI92::CheckLink
# SYNOPSIS
#       TI92::CheckLink
# DESCRIPTION
#       Check whether the TI-92 can be contacted. Like all other commands 
#       this generates an error if the TI-92 cannot be contacted, so use
#       'catch'.
#
#---------------------------------------------------------------------------
#
# NAME
#       TI92::ReceiveFile92
# SYNOPSIS
#       TI92::ReceiveFile92 source target comment cmd
# DESCRIPTION
#       Receive variable `source' from the TI-92, and store it to file 
#       `target'. If `comment' is nonempty it will be the comment of the
#       created TI-Graph Link file `target'. If `target' is empty, a target
#       name will be generated of the source name in the form "varname.92x". 
#       If `comment' is empty, an appropriate comment will be generated in a 
#       TI-Graph Link like manner. If `cmd' is nonempty it will be evaluated 
#       periodically, allowing progress indicators to display. 'cmd' will get
#       the number of bytes already received and the total number of bytes to
#       be received concatenated. 'source' should be a complete variable name
#       in the form "folder\file". Make sure you use double-backslashs when 
#       necessary to keep the backslash substitution from crippeling your 
#       variable name.
# EXAMPLE
#       # update Tk label containing receive progress in percent
#       proc DisplayProgress {numReceived total} {
#           .progress configure -text "[expr {$numReceived*100 / $total}]%"
#       }
#       TI92::ReceiveFile "main\\myvar" "" "" DisplayProgress
#
#---------------------------------------------------------------------------
#
# NAME
#       TI92::SendFile92
# SYNOPSIS
#       TI92::SendFile92 source target cmd
# DESCRIPTION
#       Send TI-Graph Link file `source' to the calculator where it will get
#       the name `target'. If `target' is empty, it will be read from the 
#       header of the TI-Graph Link file. If `cmd' is nonempty it is evaluated
#       periodically to allow progress indicators to display. It will get the
#       number of bytes already send and the total number of bytes to send
#       concatenated. If `target' is specified it should have the form 
#       "folder\file". Make sure you use double-backslashs when necessary to
#       keep the backslash substitution from crippeling your variable name.
# EXAMPLE
#       # update Tk label containing receive progress in percent
#       proc DisplayProgress {numSend total} {
#           .progress configure -text "[expr {$numSend*100 / $total}]%"
#       }
#       TI92::SendFile "~/ti92/myprgm.92p" "" DisplayProgress
#
#---------------------------------------------------------------------------
#
# NAME
#       TI92::ReceiveBackup
# SYNOPSIS
#       TI92::ReceiveBackup target comment cmd
# DESCRIPTION
#       Receive a backup from the calculator and store it into a TI-Graph Link
#       backup file (*.92b). If `comment' is nonempty it will be used as 
#       comment for the file. If it is empty an appropriate comment will be
#       generated. If `cmd' is nonempty it will be evaluated periodically
#       allowing progress indicators to display. It gets the number of bytes
#       already received concatenated (In oppsite to `ReceiveFile92' it is
#       impossible to determine how many bytes will be totaly received).
# EXAMPLE
#       # update Tk label containing the number of bytes received
#       proc DisplayBackupProgress {numReceived} {
#           .progress configure -text "$numReceived bytes"
#       }
#       TI92::ReceiveBackup "./backup.92b" "" DisplayBackupProgress
#
#---------------------------------------------------------------------------
#
# NAME
#       TI92::SendBackup
# SYNOPSIS
#       TI92::SendBackup backupFile cmd
# DESCRIPTION
#       Send a backup file `backupFile' to the calculator. If `cmd' is 
#       nonempty it will be evaluated periodically allowing progress 
#       indicators to display. It will get the number of bytes already send 
#       and the total number of bytes to be send concatenated.
# EXAMPLE
#       # update Tk label containing send progress in percent
#       proc DisplayProgress {numSend total} {
#           .progress configure -text "[expr {$numSend*100 / $total}]%"
#       }
#       TI92::SendBackup "./backup.92b" DisplayProgress
#
#---------------------------------------------------------------------------
#
# NAME 
#       TI92::ReceiveFS
# SYNOPSIS
#       TI92::ReceiveFS cmd
# DESCRIPTION
#       Receive the calculator's file system and return it as a list:
#           {folder1 folder2 ...}
#       `folderX' is yet another list:
#           {variable1 variable2 ...}
#       `variableX' is a list, too:
#           {varName varType varSize}
#       `varName' is the variable's name, `varType' is the number of the type
#       of the variable (e.g. 18 for PRGM, 12 for STR) and `varSize' is the
#       variable's size as it is displayed at the var-link screen. You may use
#       the array `TI92::VarTypeName' to convert the type-numbers into 
#       type-strings, e.g. `$TI92::VarTypeName(18)' returns `PRGM' and 
#       `$TI92::VarTypeName(12)' returns `STR'. `TI92::VarTypeExt' is a 
#       simimilar array that contains the filename extensions for each type, 
#       e.g. `$TI92::VarTypeName(18)' returns `.92p'. 
#       If your calculator contains the following variables:
#
#           GAMES
#               gorilla     PRGM    3425
#               gorillhs    PIC       19
#               gorillwt    PIC       14
#               nibbles     PRGM    7582
#           MAIN
#               y1          FUNC      26
#               y2          FUNC      24
#
#       The list returned by TI92::ReceiveFS will look like this:
#           {{games {{gorilla 18 3425} {gorillhs 16 19} {gorillwt 16 14}}}
#            {main {{y1 19 26} {y2 19 24}}}}
#       If `cmd' is nonempty, it will be evaluated periodically to allow 
#       progress indicators to display. It may also be used to directly 
#       implement some kind of `ls' command. It will get 3 arguments 
#       concatenated: the name of the variable just received, its type and
#       its size.
# EXAMPLE:
#       # display progress of file system receiving
#       proc DisplayFSProgress {name type size} {
#           global Count
#           puts -nonewline "\receiving... got [incr Count] entries"
#       }
#       set Count 0
#       set fs [TI92::ReceiveFS DisplayFSProgress]
#       puts ""
#       # display the filesystem in a var-link like manner
#       foreach fold $fs {
#           set foldName [lindex $fold 0]
#           set fileList [lindex $fold 1]
#           puts [format "%-8s" [string toupper $foldName]]
#           foreach file $fileList {
#               set fileName [lindex $file 0]
#               set fileType [lindex $file 1]
#               set fileSize [lindex $file 2]
#               set fileTypeName $TI92::VarTypeName($fileType)
#               puts [format "\t%-8s %-4s %5i" \
#                       $fileName $fileTypeName $fileSize]
#           }
#       }
#
#---------------------------------------------------------------------------
#
# NAME
#       TI92::ReceiveScreen
# SYNOPSIS
#       TI92::ReceiveScreen PBMfile comment cmd
# DESCRIPTION
#       Receive a screen dump from the calculator and store it into a Portable
#       Bitmap named `PBMfile'. If `comment' is nonempty it is written to the 
#       header of that file. Else an appropriate comment is generated. If 
#       `cmd' is nonempty it is evaluated periodically to allow progress
#       indicators to display. It gets the number of bytes already received
#       and the total number of bytes to be received concatenated. The 
#       resulting file will be a type `P4' binary encoded Portable Bitmap with
#       a size of 240x128 pixels.
# EXAMPLE
#       # update Tk label containing receive progress in percent
#       proc DisplayProgress {numReceived total} {
#           .progress configure -text "[expr {$numReceived*100 / $total}]%"
#       }
#       TI92::ReceiveScreen "screendump.pbm" "" DisplayProgress
#       # display the screendump...
#       exec xli screendump.pbm
#
#---------------------------------------------------------------------------
#
# NAME
#       TI92::SendKey
# SYNOPSIS
#       TI92::SendKey keycode
# DESCRIPTION
#       Send a keycode to the calculator, just as if the user had pressed that
#       key. Valid keycodes are those described in the manual, which are 
#       returned by the `getKey()' function. This command may be used for
#       remote control...
#
#---------------------------------------------------------------------------
#
# For more information about lower level link routines, watch the source
# code, it should be well documented...
#

if {[info tclversion] < 8.0} {
    error "'ti92.tcl' requires at least Tcl version 8.0" \
          "While parsing 'ti92.tcl'"
}

namespace eval TI92 {

namespace export OpenPort ClosePort CheckLink \
                 ReceiveFile92 SendFile92 \
                 ReceiveBackup SendBackup \
                 ReceiveFS ReceiveScreen \
		 SendKey

variable TimeOut 200;      # period before timeout in 1/100 seconds
variable MaxData 512;      # maximum number of bytes to send/receive between 
                           # evaluation of callback scripts
variable BackupPckSize 4096; # size of backup packets send
variable VarTypeNames;     # names of variable types
variable VarTypeExt;       # file extensions to use for variable types
variable PortHdl "";       # handle of the port the TI-92 is connected to
variable IgnoreCheckSum 1; # ignore checksums when possible (increases speed)
variable CheckSum 0;       # checksum which is altered by `Send' and `Receive'
variable Debug 0;          # display debugging output?

##
# Names of variable types
#
array set VarTypeName {
    10 DATA
     0 EXPR
    14 FIG
    19 FUNC
    13 GDB
     4 LIST
    20 MAC
     6 MAT
    16 PIC
    18 PRGM
    12 STR
    11 TEST
    29 BACKUP
    31 FOLDER
    256 SYS
} 

##
# File extensions for each variable type
#
array set VarTypeExt {
    10 .92d
     0 .92e
    14 .92f
    19 .92f
    13 .92d
     4 .92l
    20 .92a
     6 .92m
    16 .92i
    18 .92p
    12 .92s
    11 .92t
    30 .92b
    256 .92_
}

##
# Convert a zero-terminated string into a Tcl string.
#
proc ZString {s} {
    set zero [string first "\0" $s]
    if {$zero >= 0} {
        return [string range $s 0 [expr {$zero-1}]]
    } else {
        return $s
    }
}

##
# Calculate a word-checksum of 'data'
#
proc CheckSum {data} {
    set sum 0
    binary scan $data c* bytes
    foreach byte $bytes {
        incr sum [expr {$byte&0xFF}]
    }
    return [expr {$sum&0xFFFF}]
}

##
# Return the current date in the format used in the comments of TI-Graph Link
# files.
#
proc GetDate {} {
    return [clock format [clock seconds] -format "%m/%d/%y, %H:%M"]
}

##
# Split a TI-92 variable name ("folder\file") into a list {folder file}.
# Cause an exception if the name is invalid (currently any variable containing
# special characters, e.g. greek letters, will cause an error).
#
proc SplitVarName {name} {
    set splName [split $name "\\"]
    set fold [lindex $splName 0]
    set file [lindex $splName 1]
    if {[llength $splName] != 2 || 
            [string length $fold] > 8 || [string length $fold] == 0 || 
            [string length $file] > 8 || [string length $file] == 0 ||
            ![regexp {[a-zA-Z][a-zA-Z0-9_]*\\[a-zA-z][a-zA-Z0-9-]*} $name]} {
        error "Bad variable name: '$name' (should be 'folder\\file')\
               \nNote that special characters in variable names are\
               currently not supported."
    }
    return $splName
}

##
# Open serial port device port. for communication to the TI-92.
# This routine has to be called before any other communication routines
# are used.
#
proc OpenPort {port} {
    variable PortHdl

    set PortHdl [open $port r+]
    fconfigure $PortHdl -blocking 0 -translation binary  -encoding binary \
	    -buffering full

    # this will only work if the device is a serial port 
    catch {fconfigure $PortHdl -mode 9600,n,8,1}

    # for some reason a delay is necessary here...
    after 800
}

##
# Close the serial port, previously opened by `TI92::OpenPort'.
#
proc ClosePort {} {
    variable PortHdl
    close $PortHdl
}

##
# Hacked flush: Even in blocking mode `flush' doesn't block. So I'm using this
# routine.
#
proc HackedFlush {{len 0}} {
    variable PortHdl

    after [expr {$len*2 + 1000}] {set SendDone 1}
    fileevent $PortHdl writable {
	fileevent $TI92::PortHdl writable {}
	set SendDone 1
    }
    vwait SendDone
    after cancel {set SendDone 1}
}

##
# Send data 'data' to the TI-92
#
proc Send {data} {
    variable Debug

    variable PortHdl 
    variable CheckSum

    if {$Debug} {
	binary scan $data H* hex
	puts -nonewline "< $hex"; update
    }

    fconfigure $PortHdl -blocking 0
    puts -nonewline $PortHdl $data
    incr CheckSum [CheckSum $data]
    flush $PortHdl
    HackedFlush [string length $data]
}

##
# Receive 'num' bytes from the TI-92. If `exact' is true, generate a timeout
# error if no bytes are received during `TimeOut*100' milliseconds.
# Return the bytes received.
#
proc Receive {num {exact 1}} {
    variable PortHdl 
    variable TimeOut
    variable CheckSum
    variable IgnoreCheckSum
    variable Debug

    fconfigure $PortHdl -blocking 0

    for {set data {}; set count 0; set dataLen 0} {$dataLen < $num} {} {
        if {$count >= $TimeOut && $exact} {
            error "Timeout" "While waiting for data from the TI-92"
        }
        flush $PortHdl
        set r [read $PortHdl [expr {$num-$dataLen}]]
        set rLen [string length $r]
        if {$rLen} {    # some data received
            append data $r; incr dataLen $rLen
	    if {!$IgnoreCheckSum} {
		set CheckSum [expr {($CheckSum+[CheckSum $r])&0xFFFF}]
	    }
            set count 0
        } else {        # no data received: sleep a little, incr timeout count
	    if {!$exact} break
            after 10
            incr count
        }
    }

    if {$Debug} {
	binary scan $data H* hex
	puts -nonewline "> $hex"; update  
    }

    return $data
}

##
# Clear all linkport data that may be somewhere in buffers...
#
proc ResetLink {} {
    variable PortHdl

    fconfigure $PortHdl -blocking 0
    flush $PortHdl
    read $PortHdl
}

##
# Send a packet to the TI-92. Only packets 0x06 (HEADER), 0x15 (DATA) and
# 0xA2 (EXT) have a data field and come with a checksum. For other packets
# 'data' is ignored.
# If 'script' is nonempty and the packet has a data field, 'script' is 
# evaluated whenever 'MaxData' bytes have been send and at the end of 
# transfer. It gets the number of bytes transfered and the total number 
# of bytes to transfer concatenated:
#     eval $script $numBytesTransfered $numBytes
#
proc SendPacket {command data script} {
    variable MaxData
    variable CheckSum

    set dataLen [string length $data]
    Send [binary format ccs 0x09 $command $dataLen]

    if {$command == 0x06 || $command == 0x15 || $command == 0xA2} {
	set CheckSum 0
        for {set numTransfered 0} {$numTransfered < $dataLen} {} {
	    set num [expr {$numTransfered+$MaxData < $dataLen ?
	            $MaxData : $dataLen - $numTransfered}]
	    binary scan $data "@${numTransfered}a${num}" part
	    Send $part
	    incr numTransfered $num
	    if {[string compare $script ""]} {
		eval $script $numTransfered $dataLen
	    }
        }
        Send [binary format s $CheckSum]
    }


}

##
# Receive a packet from the TI-92. Return a list of the following format: 
#     {<command> <data-length> <data>}
# If 'script' is nonempty, it is evaluated whenever 'MaxData' bytes 
# have been received and at the end of transfer. It gets the number  of 
# data-bytes received and the total number of bytes to receive concatenated:
#     eval $script $numBytesReceived $numBytes
#
proc ReceivePacket {script} {
    variable MaxData 
    variable IgnoreCheckSum
    variable CheckSum

    binary scan [Receive 4] ccs device command dataLen
    set command [expr $command&0xFF]
    set device [expr $device&0xFF]
    set data ""

    # receive data; only 0x06 (HEADER), 0x15 (DATA) and 0xA2 (EXT) contain
    # data!
    if {$command == 0x06 || $command == 0x15 || $command == 0xA2} {
	set CheckSum 0
        for {set numReceived 0} {$numReceived < $dataLen} {} {
            set num [expr {$numReceived+$MaxData < $dataLen ?
		    $MaxData : $dataLen-$numReceived}]
            append data [Receive $num]
            incr numReceived $num
            if {[string compare $script ""]} {
                eval $script $numReceived $dataLen
            }
        }
	set calcCSum $CheckSum
        binary scan [Receive 2] s packetCSum
        if {!$IgnoreCheckSum && ($packetCSum&0xFFFF) != $calcCSum} {
            SendPacket 0x5A "" ""
            error "Received bad checksum\
		   (p:$packetCSum != c:$calcCSum)" \
		   "While receiving data from the TI-92"
        }
    }

    if {$device != 0x89} {
        error "Device didn't identify as a TI-92" \
              "While receiving data from the TI-92"
    }

    return [list $command $dataLen $data]
}

##
# Send an Acknoledge to the TI-92
#
proc SendAck {} {
    SendPacket 0x56 "" ""
}

##
# Get an Acknoledge from the TI-92.
#
proc ReceiveAck {} {
    set packet [ReceivePacket ""]
    if {[lindex $packet 0] != 0x56} {
        error "Invalid response from TI-92" \
  	      "Received {$packet} while waiting for ACK"
    }
}

##
# Send a Request For Data to the TI-92.
#
proc SendRFD {} {
    SendPacket 0x09 "" ""
}

##
# Receive a Request For Data from the TI-92.
#
proc ReceiveRFD {} {
    set packet [ReceivePacket ""]   
    if {[lindex $packet 0] != 0x09} {
        error "Invalid response from TI-92" \
	      "Received {$packet} while waiting for RFD"
    }
}

##
# Send End Of Transmission to the TI-92.
#
proc SendEOT {} {
    SendPacket 0x92 "" ""
}

##
# Get and End Of Transmission from the TI-92.
#
proc ReceiveEOT {} {
    set packet [ReceivePacket ""]
    if {[lindex $packet 0] != 0x92} {
        error "Invalid response from TI-92" \
	      "Received {$packet} while waiting for EOT"
    }
}

##
# Send a Checksum Error packet to the TI-92
#
proc SendCERR {} {
    SendPacket 0x5A "" "" 
}

##
# Check whether the calculator replies
#
proc CheckLink {} {
    ResetLink
    SendPacket 0x68 0 ""
    ReceiveAck
}

##
# Send a variable to the TI-92. 'var' are the raw variable data, as
# they're stored on the calc. 'name' is the target name "folder\file".
# If script is nonempty, it is evaluated periodically (see 'MaxData'). 
# It gets the number of bytes transfered and the total number of bytes 
# to transfer concatenated (Note that the total number of bytes is a 
# little greater than the variable size):
#     eval $script $numBytesTransfered $numBytes
#
proc SendRawVar {name type var script} {
    set varLen [string length $var]
    set nameLen [string length $name]
    ResetLink
    SendPacket 0x06 \
            [binary format sxxcca* $varLen $type $nameLen $name] ""
    ReceiveAck
    ReceiveRFD
    SendAck
    SendPacket 0x15 [binary format x4a* $var] $script
    ReceiveAck
    SendEOT
    ReceiveAck
}

##
# Send a raw backup to the TI-92. 'romVersion' is the ROM-version of the
# backup, 'backup' are the raw backup data. If 'script' is nonempty it may
# be used for progress indicators. It will be evaluated whenever 1K was 
# received, and gets the number of bytes transfered and the total number of 
# bytes to transfer concatenated.
#
proc SendRawBackup {romVersion backup script} {
    variable BackupPckSize

    set romVLen [string length $romVersion]
    ResetLink
    set len [string length $backup]

    SendPacket 0x06 [binary format "sscca*" \
	    $BackupPckSize 0 0x1D $romVLen $romVersion] ""
    ReceiveAck

    for {set I 0} {$I < $len} {} {
        SendPacket 0x06 [binary format "sscca*" $BackupPckSize 0 0x1D \
                        [string length $romVersion] $romVersion] ""
        ReceiveAck
        ReceiveRFD
        SendAck
        set blockLen [expr \
		{$I+$BackupPckSize < $len ? $BackupPckSize : $len-$I}]
        binary scan $backup "@${I}a${blockLen}" dataBlock
        incr I $blockLen
        SendPacket 0x15 $dataBlock ""
        ReceiveAck
        
        if {[string compare $script ""]} {
            eval $script $I $len
        }
    }

    SendEOT
}

##
# Receive file system information from the TI-92. The return value is a
# list. Each element of that list is another list that describes a folder in
# the following way:
#     {folder-name {filelist}}
# Filelist is another list, whose elements have the following format:
#     {file-name file-type file-size}
# If 'cmd' is nonempty, it specifies a script that is evaluated whenever
# a file/folder is received. It gets 3 arguments concatenated: 
#     eval $cmd $name $type $num
# 'name' is the name of the file/folder that was read, 'type' is its type
# (0x1F for folders) and 'num' is the file-size for files or the number of
# files in a folder. You may use array 'VarTypeName' for getting descriptions
# for each numberic variable-type.
#
proc ReceiveFS {cmd} {
    ResetLink

    # send request for receiving file system
    SendPacket 0xA2 [binary format x4cc 0x19 0x00] ""
    ReceiveAck
    set header [ReceivePacket ""]
    SendAck
    
    for {set end 0; set folderList {}} {!$end} {} {

        # receive DATA packet
        SendRFD
        ReceiveAck
        set fsDataPacket [ReceivePacket ""]
        if {[lindex $fsDataPacket 0] != 0x15} {
            error "Invalid response from TI-92" \
    	          "Received {$fsDataPacket} while waiting for DATA during FS transfer"
        }
        SendAck 

        # process DATA packet
        binary scan [lindex $fsDataPacket 2] x4A8ccsx2 name type _duno_ num
        set name [ZString $name]
        if {$type == 0x1F} {   # folder
            set folderName $name
            set numFiles $num
            set fileList {}
        } else {               # file
            lappend fileList [list $name $type $num]
        }
        if {[llength $fileList] == $numFiles} {
            lappend folderList [list $folderName $fileList]
        }

        # evaluate "callback" script
        if {[string compare $cmd ""]} {
            eval $cmd $name $type $num
        }

        # receive EOT or File System Continuation packet 
        set command [lindex [set packet [ReceivePacket ""]] 0]
        if {$command == 0x92} {
            set end 1
        } elseif {$command != 0x78} {
            error "Invalid response from TI-92" \
    	          "Received {$packet} while waiting for EOT or FS-CONT during FS transfer"
	}
        SendAck
    }

    return $folderList
}

##
# Receive variable 'name' from the TI-92. The return value is a list:
#     {type varLen var}
# 'var' are the variable data as they're stored on the TI-92. 'varLen'
# is its length, as displayed by the Var-Link Screen. 
# If 'script' is nonempty, it is evaluated periodically (see 'MaxData').
# It gets two arguments concatenated, the number of bytes received and 
# the total number of bytes to receive:
#     eval $script $numBytesReceived $numBytes
#
proc ReceiveRawVar {name script} {
    ResetLink

    # send request for receiving file 'name'
    SendPacket 0xA2 [binary format x4cca* 0x05 [string length $name] $name] ""
    ReceiveAck

    # receive HEADER packet
    set header [ReceivePacket ""]
    if {[lindex $header 0] != 0x06} {
        error "Invalid response from TI-92" \
	      "Received {$header} while waiting for variable HEADER"
    }
    binary scan [lindex $header 2] sxxcca* varLen type nameLen name
    SendAck

    # receive DATA packet
    SendRFD
    ReceiveAck
    set data [ReceivePacket $script]

    if {[lindex $data 0] != 0x15} {
        error "Invalid response from TI-92" \
	      "Received {$data} while waiting for variable DATA"
    }
    binary scan [lindex $data 2] x4a* var
    SendAck

    # end of transmission
    ReceiveEOT
    SendAck

    return [list $type $varLen $var]
}

##
# Receive a backup from the calculator. Return a list:
#     {ROM-version backupLen backup}
# where 'backup' are the raw binary backup data and 'ROM-Version' is a string
# describing the ROM-version of the calc. 
# If 'script' is nonempty it is evaluated each time 1K is received and will
# get the number of bytes received concatenated.
#
proc ReceiveRawBackup {script} {
    ResetLink

    # send request for receiving backup
    SendPacket 0xA2 [binary format x4cca* 0x1D 11 "main\\backup"] ""
    ReceiveAck

    # receive backup in 1K parts
    set backup ""; set I 0
    while {1} {

        # receive header or EOT
        set header [ReceivePacket ""]
        if {[lindex $header 0] == 0x92} {  # End Of Transmission
            break
        }
        if {[lindex $header 0] != 0x06} {
            error "Invalid response from TI-92" \
	          "Received {$header} while waiting for backup HEADER"
        } 
        SendAck
        binary scan [lindex $header 2] x6a* romVersion

        # receive data packet
        SendRFD 
        ReceiveAck
        set data [ReceivePacket ""]
        if {[lindex $data 0] != 0x15} {  # check for data packet
            error "Invalid response from TI-92" \
	          "Received {$data} while waiting for variable DATA"
        }
        SendAck
        binary scan [lindex $data 2] x4a* dataBlock
        append backup $dataBlock
        incr I [string length $dataBlock]
        if {[string compare $script ""]} {
	    eval $script $I
        }
    }

    SendAck
    return [list $romVersion $I $backup]
}

##
# Receive screen dump from the TI-92. The return value is a binary string,
# each bit representing one pixel of the 240x128 screen pixels.
# If 'cmd' is nonempty, it is evaluated periodically (see 'MaxData').
# It gets two arguments concatenated, the number of bytes received and 
# the total number of bytes to receive:
#     eval $cmd $numBytesReceived $numBytes
#
proc ReceiveRawScreen {cmd} {
    variable IgnoreCheckSum 
    global errorInfo

    ResetLink

    # send request for screen dump
    SendPacket 0x6D "" ""
    ReceiveAck

    # receive screen dump DATA packet; checksums are generated differently for
    # screen dumps, so I disable checksum check
    set ics $IgnoreCheckSum
    set IgnoreCheckSum 1
    if {[catch {set data [ReceivePacket $cmd]} errMsg]} {
	set IgnoreCheckSum $ics
	error $errMsg $errorInfo
    } 
    set IgnoreCheckSum $ics

    if {[lindex $data 0] != 0x15} {
        error "Invalid response from TI-92" \
	      "Received {$data} while waiting for screen dump DATA"
    }
    SendAck

    return [lindex $data 2]
}

##
# Send a keycode to the calculator (for remote control). Valid keycodes are 
# those described in the manual, which are returned by `getkey()' function.
#
proc SendKey {keycode} {
    ResetLink
    Send [binary format ccs 0x09 0x87 $keycode]
    ReceiveAck
}

##
# Open and read a TI-Graph Link (*.92?) file. Return a list:
#     {varName type comment varLen var}
# 'var' are the raw variable data as they are stored on the calc and 'varLen'
# is the variable size as it's displayed in the Graph-Link screen. 'varName' is
# the name that the variable had on the calc ("folder\filename"). It may differ
# from the filename, since it stays the same even if the file is renamed.
#
proc ReadFile92 {name} {
    variable IgnoreCheckSum

    set hdl [open $name r]
    fconfigure $hdl -translation binary -encoding binary
    set file [read $hdl]
    close $hdl
    binary scan $file "a8x2A8A40x2x4A8sx2ix6a*" \
	    sig foldName comment fileName type fileLen data
    set varLen [expr {[string length $data]-2}]
    set foldName [ZString $foldName]
    set fileName [ZString $fileName]
    set comment [ZString $comment]
    binary scan $data "a${varLen}s" var checkSum 
    set checkSum [expr {$checkSum&0xFFFF}]

    if {[string compare $sig "**TI92**"]} {
        error "File '$name' does not seem to be a valid TI-Graph Link file" \
              "Signature of '$name' didn't match '**TI92**'"
    }
    if {[string length $file] != $fileLen} {
        error "File '$name' does not seem to be a valid TI-Graph Link file" \
              "Actual file size ([string length $file]) differs from \
               file length field ($fileLen)"
    }
    if {!$IgnoreCheckSum && $checkSum != [CheckSum $var]} {
	error "File '$name' does not seem to be a valid TI-Graph Link file" \
              "Bad checksum"
    }
    if {$type > 20} {
        error "File '$name' is of unknown variable type" \
              "Var-type field is $type"
    }
    return [list "$foldName\\$fileName" $type $comment $varLen $var]
}

##
# Open and read a TI-Graph Link backup (*.92b) file. Return a list:
#     {ROM-version backupLen backup}
# 'backup' is the raw backup, 'backupLen' its length.
#
proc ReadBackup {name} {
    variable IgnoreCheckSum

    set hdl [open $name r]
    fconfigure $hdl -translation binary -encoding binary
    set file [read $hdl]
    close $hdl
    binary scan $file "a8x2x8A40x2x4A8sx2ix2a*" \
	    sig comment romVersion type fileLen data
    set backupLen [expr {[string length $data]-2}]
    set romVersion [ZString $romVersion]
    set comment [ZString $comment]
    binary scan $data "a${backupLen}s" backup checkSum 
    set checkSum [expr {$checkSum&0xFFFF}]

    if {[string compare $sig "**TI92**"]} {
        error "File '$name' does not seem to be a valid TI-Graph Link backup" \
              "Signature of '$name' didn't match '**TI92**'"
    }
    if {[string length $file] != $fileLen} {
        error "File '$name' does not seem to be a valid TI-Graph Link backup" \
              "Actual file size ([string length $file]) differs from \
               file length field ($fileLen)"
    }
    if {!$IgnoreCheckSum && $checkSum != [CheckSum $backup]} {
	error "File '$name' does not seem to be a valid TI-Graph Link backup" \
              "Bad checksum"
    }
    if {$type != 29} {
        error "File '$name' is a TI-Graph Link file but not of type 'backup'" \
              "Var-type field is $type"
    }
    return [list $romVersion $backupLen $backup]
}

##
# Create (overwrite) a TI-Graph Link (*.92?) file for the variable that had
# name 'varName' ("folder\file") on the calc. 'type' is the variable's type,
# 'var' are the raw variable data. 'comment' is the file's comment and should
# not be longer than 40 characters.
#
proc WriteFile92 {name varName type comment var} {
    set splName [SplitVarName $varName]
    set fold [lindex $splName 0]
    set file [lindex $splName 1]
    set varLen [string length $var]
    set fileLen [expr {$varLen+0x58}]

    set file [binary format "a8c2a8A40c6a8sc2ic6a${varLen}s" \
            "**TI92**" {1 0} $fold $comment {1 0 0x52 0 0 0}  $file $type \
            {0 0} $fileLen {0xA5 0x5A 0 0 0 0} $var [CheckSum $var]]

    set hdl [open $name "w"]
    fconfigure $hdl -translation binary -encoding binary
    puts -nonewline $hdl $file
    close $hdl
}

##
# Create (overwrite) a TI-Graph Link backup file (*.92b) 'name' with the 
# backup-data provided by 'backup' and ROM-version 'romVersion'. 'comment' 
# is the backup's comment and should not be longer than 40 characters.
#
proc WriteBackup {name romVersion comment backup} {
    set backupLen [string length $backup]
    set fileLen [expr {$backupLen+0x54}]

    set file [binary format "a8c10A40c6a8sx2ic2a${backupLen}s" \
            "**TI92**" {1 0 0x10 0 0 0 0 0 0 0} $comment {1 0 0x52 0 0 0} \
            $romVersion 29 $fileLen {0xA5 0x5A} $backup [CheckSum $backup]]

    set hdl [open $name "w"]
    fconfigure $hdl -translation binary -encoding binary
    puts -nonewline $hdl $file
    close $hdl
}

##
# Write (overwrite) PBM (Portable Bitmap) file `name' with the screen-dump data
# provided by `pic'. `comment' is any comment that should be added to the file
# header. The resulting picture is a type `P4' binary coded file.
#
proc WritePBM {name comment pic} {
    set file [binary format "a*a2a*a1a*a*" \
	    "P4\n" "# " $comment "\n" "240 128\n" $pic]

    set hdl [open $name "w"]
    fconfigure $hdl -translation binary -encoding binary
    puts -nonewline $hdl $file
    close $hdl
}

##
# Send TI-Graph Link (*.92?) file 'srcName' to the calc. 'destName' may 
# explicitely specify the destination for to store the variable on the 
# calc ("folder\filename"). If 'destName' is an empty string, it will be
# read from the file 'srcName' (the original variable name before the file
# was transfered to the PC).
# If nonempty, the 'script' argument is evaluated periodically, you may use 
# it for status display. it gets the number of bytes transmitted and the 
# total number of bytes to transmit concatenated:
#     eval $script $numBytesTransmitted $numBytes
# For example to display the percentage:
#
#     proc DisplaySendProgress {numBytesTransmitted numBytes} {
#         puts "\r Transfer in progress... \
#                 [expr {$numBytesTransmitted*100 / $numBytes}]%"
#     }
#
#     SendFile92 "h:\fargofiles\myprog.92p" DisplaySendProgress
#
proc SendFile92 {srcName destName script} {
    set file [ReadFile92 $srcName]

    if {![string compare $destName ""]} { set destName [lindex $file 0] }
    set type [lindex $file 1]
    set var [lindex $file 4]

    SendRawVar $destName $type $var $script
}

##
# Receive a variable from the calc and store it into a TI-Graph Link variable.
# 'srcName' is the variable name on the calc ("folder\file"). 'destName' is
# the destination name. If it is empty, the file will be put into the current
# directory, named by calculator source variable ("varname.92?") where '?'
# will be chosen according to the file's type. 'comment' will be the file's 
# comment. If it's empty it will be automatically generated. Script may be 
# used for progress indicators. See 'SendFile92' for details...
#
proc ReceiveFile92 {srcName destName comment script} {
    variable VarTypeExt 
    variable VarTypeName

    set splName [SplitVarName $srcName]
    set var [ReceiveRawVar $srcName $script]
    set type [lindex $var 0]
    set varData [lindex $var 2]
    if {![string compare $destName ""]} {
        set destName \
            "[lindex $splName 1]$VarTypeExt($type)"
    }
    if {![string compare $comment ""]} {
        set comment "$VarTypeName($type) variable dated [GetDate]"
    }

    WriteFile92 $destName $srcName $type $comment $varData
}

##
# Send backup file 'name' to the TI-92. If 'cmd' is nonempty it is 
# evaluated everytime 1K was transferred. It gets the number of bytes 
# transferred and the total number of bytes to transfer concatenated:
#     eval $cmd $numBytesTransfered $numBytes
#
proc SendBackup {name cmd} {
    set b [ReadBackup $name]
    SendRawBackup [lindex $b 0] [lindex $b 2] $cmd
}

##
# Receive a backup from the TI-92 and store it in file 'name' with comment
# 'comment'. If 'comment' is empty an appropriate comment will be generated.
# If 'script' is nonempty it is evaluated whenever 1K was received. it gets 
# the number of bytes received concatenated:
#     eval $script $numBytesReceived
#
#
proc ReceiveBackup {name comment script} {
    set b [ReceiveRawBackup $script]

    if {![string compare $comment ""]} {
        set comment "Backup file dated [GetDate]"
    }

    WriteBackup $name [lindex $b 0] $comment [lindex $b 2]
}

##
# Receive a screen dump from the calculator and store it in `name' as a 
# PBM (Portable Bitmap) file. If `comment' is empty an appropriate comment 
# will be generated. If `cmd' is nonempty it is evaluated periodically to 
# allow progress indicators to display. It gets the number of bytes received
# and the total number of bytes to receive concatenated.
#
proc ReceiveScreen {name comment cmd} {
    set pic [ReceiveRawScreen $cmd]
    
    if {![string compare $comment ""]} {
	set comment "Screen dump dated [GetDate]"
    }

    WritePBM $name $comment $pic
}

} 
# namespace 'TI92' ends here





